import 'package:adt_annotation_base/adt_annotation_base.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:source_gen/source_gen.dart';

TypeChecker unionTypeChecker = TypeChecker.fromRuntime(Union);
TypeChecker tupleTypeChecker = TypeChecker.fromRuntime(Tuple);
TypeChecker recordTypeChecker = TypeChecker.fromRuntime(Record);
TypeChecker opaqueTypeChecker = TypeChecker.fromRuntime(Opaque);

TypeD typeDFromReader(ConstantReader reader) {
  final name = reader.read('name').symbolValue;
  final arguments = reader
      .read('arguments')
      .listValue
      .map(ConstantReader.new)
      .map(typeDFromReader)
      .toList();
  final assertions = reader
      .read('assertions')
      .listValue
      .map(
        (e) => e.toStringValue()!,
      )
      .toList();
  final namespace = reader.read('namespace').isNull
      ? null
      : reader.read('namespace').symbolValue;
  final equality = reader.read('equality').isNull
      ? null
      : reader.read('equality').stringValue;
  final hash =
      reader.read('hash').isNull ? null : reader.read('hash').stringValue;
  final fromJson = reader.read('fromJson').isNull
      ? null
      : reader.read('fromJson').stringValue;
  final toJson =
      reader.read('toJson').isNull ? null : reader.read('toJson').stringValue;
  if (reader.read('nullable').boolValue) {
    return TypeD.n(
      name,
      args: arguments,
      asserts: assertions,
      namespace: namespace,
      equality: equality,
      hash: hash,
      fromJson: fromJson,
      toJson: toJson,
    );
  }
  return TypeD(
    name,
    args: arguments,
    asserts: assertions,
    namespace: namespace,
    equality: equality,
    hash: hash,
    fromJson: fromJson,
    toJson: toJson,
  );
}

Union unionFromReader(ConstantReader reader) {
  return Union(
    reader.read('body').mapValue.map((caseName, caseBody) => MapEntry(
          ConstantReader(caseName).symbolValue,
          caseBody!.toMapValue()!.map((memberName, memberType) => MapEntry(
                ConstantReader(memberName).symbolValue,
                typeDFromReader(ConstantReader(memberType!)),
              )),
        )),
    structuralImplements: reader.read('structuralImplements').boolValue,
    deriveMode: UnionVisitDeriveMode
        .values[reader.read('deriveMode').read('index').intValue],
    topLevel: reader.read('topLevel').boolValue,
    deriveNamed: reader.read('deriveNamed').boolValue,
    caseMethods: reader
        .read('caseMethods')
        .mapValue
        .map((caseName, caseBody) => MapEntry(
              ConstantReader(caseName).symbolValue,
              methodsFromReader(ConstantReader(caseBody)),
            )),
  );
}

Methods methodsFromReader(ConstantReader reader) {
  return Methods(
    static:
        reader.read('static').mapValue.map((methodName, function) => MapEntry(
              ConstantReader(methodName).symbolValue,
              AnnotatedFunction.make(
                function!.toFunctionValue()! as FunctionElement,
              ),
            )),
    instance:
        reader.read('instance').mapValue.map((methodName, function) => MapEntry(
              ConstantReader(methodName).symbolValue,
              AnnotatedFunction.make(
                function!.toFunctionValue()! as FunctionElement,
              ),
            )),
    get: reader.read('get').mapValue.map((attributeName, function) => MapEntry(
          ConstantReader(attributeName).symbolValue,
          AnnotatedFunction.make(
            function!.toFunctionValue()! as FunctionElement,
          ),
        )),
    set: reader.read('set').mapValue.map((attributeName, function) => MapEntry(
          ConstantReader(attributeName).symbolValue,
          AnnotatedFunction.make(
            function!.toFunctionValue()! as FunctionElement,
          ),
        )),
  );
}

Tuple tupleFromReader(ConstantReader reader) {
  return Tuple(
    reader
        .read('body')
        .listValue
        .map(ConstantReader.new)
        .map(typeDFromReader)
        .toList(),
  );
}

Opaque opaqueFromReader(ConstantReader reader) {
  return Opaque(
    typeDFromReader(reader.read('type')),
    exposeConstructor: reader.read('exposeConstructor').boolValue,
  );
}

Record recordFromReader(ConstantReader reader) {
  return Record(
    reader.read('body').mapValue.map((name, type) => MapEntry(
          ConstantReader(name).symbolValue,
          typeDFromReader(ConstantReader(type)),
        )),
    deriveMode: RecordConstructorDeriveMode
        .values[reader.read('deriveMode').read('index').intValue],
  );
}

DataExpr dataExprFromReader(ConstantReader reader) {
  if (reader.instanceOf(unionTypeChecker)) {
    return unionFromReader(reader);
  }
  if (reader.instanceOf(tupleTypeChecker)) {
    return tupleFromReader(reader);
  }
  if (reader.instanceOf(recordTypeChecker)) {
    return recordFromReader(reader);
  }
  if (reader.instanceOf(opaqueTypeChecker)) {
    return opaqueFromReader(reader);
  }
  throw TypeError();
}

TypeParam typeParamFromReader(ConstantReader reader) {
  return TypeParam(
    reader.read('name').symbolValue,
    typeDFromReader(reader.read('constraint')),
  );
}

class AnnotatedFunction implements Function {
  final FunctionElement _element;

  AnnotatedFunction(this._element);
  dynamic call() {
    return null;
  }

  static Expando _expando = Expando();
  static FunctionElement getElement(Function closure) =>
      (_expando[closure] as AnnotatedFunction)._element;
  static void Function() make(FunctionElement element) {
    final closure = () => {};
    final data = AnnotatedFunction(element);
    _expando[closure] = data;
    return closure;
  }
}

DataAnnotation dataAnnotationFromReader(ConstantReader reader) =>
    DataAnnotation(
        reader.read('name').symbolValue,
        reader
            .read('typeParams')
            .listValue
            .map(ConstantReader.new)
            .map(typeParamFromReader)
            .toList(),
        dataExprFromReader(reader.read('body')),
        mixin: reader
            .read('mixin')
            .listValue
            .map(ConstantReader.new)
            .map(typeDFromReader)
            .toList(),
        implements: reader
            .read('implements')
            .listValue
            .map(ConstantReader.new)
            .map(typeDFromReader)
            .toList(),
        deriveEquality: reader.read('deriveEquality').boolValue,
        deriveRuntimeType: reader.read('deriveRuntimeType').boolValue,
        deriveToString: reader.read('deriveToString').boolValue,
        deriveCopyWith: reader.read('deriveCopyWith').boolValue,
        deriveToJson: reader.read('deriveToJson').boolValue,
        deriveFromJson: reader.read('deriveFromJson').boolValue,
        methods: methodsFromReader(reader.read("methods")));

final typeTypeChecker = TypeChecker.fromRuntime(Type);
