import 'package:adt_annotation_base/adt_annotation_base.dart';
import 'package:adt_generator/src/reviving.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/dart/element/type.dart';

String stringLowercasedWithoutUnderlines(String s) => s.startsWith('_')
    ? stringLowercasedWithoutUnderlines(s.substring(1))
    : s[0].toLowerCase() + s.substring(1);

String unqualifiedStringFromSymbol(Symbol s) {
  final asString = s.toString();
  return asString.split('"')[1];
}

String emptyOrSurrounded(Iterable<String> strings, [String seps = '<,>']) =>
    strings.isEmpty ? '' : '${seps[0]}${strings.join(seps[1])}${seps[2]}';

String implementsToCode(Iterable<TypeD> implements) => implements.isEmpty
    ? ""
    : "implements ${implements.map((t) => t.toCode()).join(', ')}";

class Signature {
  final String returnType;
  final String name;
  final List<String> positionalTypes;
  final List<String> positionalNames;
  final Map<String, String> named;

  Signature(
    this.returnType,
    this.name,
    this.positionalTypes,
    this.positionalNames,
    this.named,
  );
}

String dartTypeToString(DartType t) =>
    t.getDisplayString(withNullability: true).replaceAll('*', '');

Signature signatureFromAnnotatedFunction(FunctionElement element) {
  return Signature(
    element.returnType.isDynamic ? "" : dartTypeToString(element.returnType),
    element.name,
    element.type.normalParameterTypes.map(dartTypeToString).toList(),
    element.type.normalParameterNames,
    element.type.namedParameterTypes.map(
      (key, value) => MapEntry(
        key,
        dartTypeToString(value),
      ),
    ),
  );
}

String signatureToCallCode(Signature signature, [bool replaceSelf = false]) =>
    '''${signature.name}(${signature.positionalNames.map((e) => replaceSelf && e == "self" ? "this" : e).followedBy(signature.named.keys.map((e) => '$e: $e')).join(', ')})''';
String signatureToParamsCode(Signature signature, [bool replaceSelf = false]) {
  final res = StringBuffer();
  for (var i = 0; i < signature.positionalNames.length; i++) {
    final name = signature.positionalNames[i];
    final type = signature.positionalTypes[i];
    if (replaceSelf && name == "self") {
      continue;
    }
    res.write(type);
    if (type.isNotEmpty) {
      res.write(' ');
    }
    res.write(name);
    res.write(', ');
  }
  if (signature.named.isNotEmpty) {
    res.write('{');
    res.writeAll(
        signature.named.entries.map((e) => '${e.value} ${e.key}'), ", ");
    res.write('}');
  }
  return res.toString();
}

String getterToCode(Symbol name, Function function) {
  final sig =
      signatureFromAnnotatedFunction(AnnotatedFunction.getElement(function));
  assert(sig.positionalNames[0] == "self");
  final call = signatureToCallCode(sig, true);
  return '''${sig.returnType} get ${name.toCode()} => $call;''';
}

String setterToCode(Symbol name, Function function) {
  final sig =
      signatureFromAnnotatedFunction(AnnotatedFunction.getElement(function));
  assert(sig.positionalNames[0] == "self");
  final call = signatureToCallCode(sig, true);
  return '''set ${name.toCode()}(${signatureToParamsCode(sig, true)}) => $call;''';
}

String instanceMethodToCode(Symbol name, Function function) {
  final sig =
      signatureFromAnnotatedFunction(AnnotatedFunction.getElement(function));
  assert(sig.positionalNames[0] == "self");
  final call = signatureToCallCode(sig, true);
  return '''${sig.returnType} ${name.toCode()}(${signatureToParamsCode(sig, true)}) => $call;''';
}

String staticMethodToCode(Symbol name, Function function) {
  final sig =
      signatureFromAnnotatedFunction(AnnotatedFunction.getElement(function));
  final call = signatureToCallCode(sig);
  return '''static const ${name.toCode()} = ${sig.name};''';
}

String methodsToCode(Methods methods) {
  final getters =
      methods.get.entries.map((e) => getterToCode(e.key, e.value)).join('\n');
  final setters =
      methods.set.entries.map((e) => setterToCode(e.key, e.value)).join('\n');
  final instance = methods.instance.entries
      .map((e) => instanceMethodToCode(e.key, e.value))
      .join('\n');
  final static = methods.static.entries
      .map((e) => staticMethodToCode(e.key, e.value))
      .join('\n');
  return [getters, setters, instance, static].join('\n');
}

extension aaaaa on DataAnnotation {
  String typeParamsToCode() => typeParams.toCode();
  String parameterizedTypeToCode() => name.toCode() + typeParamsToCode();
  String instantiationToCode() => instantiationToType().toCode();
  TypeD instantiationToType() => TypeD(
        name,
        args: typeParams.toInstantiation().toList(),
      );
}

extension aaaa on Symbol {
  String toCode() => unqualifiedStringFromSymbol(this);
}

extension aaa on TypeD {
  String toCode() =>
      (namespace == null ? '' : '${namespace!.toCode()}.') +
      name.toCode() +
      emptyOrSurrounded(arguments.map((e) => e.toCode())) +
      (nullable ? '?' : '');
  String toTypeLiteralCode() => nullable
      ? 'Nullable<${T(name, args: arguments, asserts: assertions, namespace: namespace).toCode()}>'
      : toCode();
}

String bodyToFields(Map<Symbol, TypeD> body) => body.entries
    .map((e) => '''final ${e.value.toCode()} ${e.key.toCode()};''')
    .join('\n');

String initializerArgsFromSymbols(
  Iterable<Symbol> body, {
  bool required = false,
}) =>
    body
        .map((e) => '${required ? "required " : ""}this.${e.toCode()}')
        .join(', ');

String initializerAssertionsFromBody(Map<Symbol, TypeD> body) {
  final r = body.entries
      .expand((e) => assertStatementsFromTypeD(e.value, e.key.toCode()))
      .map((e) => '$e,')
      .join('\n');
  return r;
}

String maybeGenerate(bool condition, String Function() generate) =>
    condition ? generate() : '';

String _hashToCode(MapEntry<Symbol, TypeD> e) {
  const defaultHash = '{}';
  final hash = e.value.hash ?? defaultHash;
  return hash.replaceAll('{}', e.key.toCode());
}

Iterable<String> hashablePartsFrom(
  TypeD instantiatedType,
  Map<Symbol, TypeD> body,
) =>
    [
      '(${instantiatedType.name.toCode()})',
      ...instantiatedType.arguments.map((e) => e.toTypeLiteralCode()),
      ...body.entries.map(_hashToCode),
    ];

String hashExpressionFromTypeAndBody(
  TypeD instantiatedType,
  Map<Symbol, TypeD> body,
) =>
    hashExpressionFromHashable(hashablePartsFrom(instantiatedType, body));
String hashExpressionFromHashable(Iterable<String> hashable) {
  if (hashable.length == 1) {
    return '${hashable.single}.hashCode';
  }
  if (hashable.length <= 20) {
    return 'Object.hash(${hashable.join(', ')})';
  }

  return 'Object.hashAll([${hashable.join(', ')}])';
}

String _equalityToCode(MapEntry<Symbol, TypeD> e) {
  const defaultEquality = '{a} == {b}';
  final equality = e.value.equality ?? defaultEquality;
  return equality
      .replaceAll('{a}', 'this.${e.key.toCode()}')
      .replaceAll('{b}', 'other.${e.key.toCode()}');
}

String _specificEqualityExpressionFor(String type, Map<Symbol, TypeD> body) {
  return '(' +
      'other is $type &&\n' +
      [
        'true',
        ...body.entries.map(_equalityToCode),
      ].join(' &&\n') +
      ')';
}

String equalityExpressionFor(String type, Map<Symbol, TypeD> body) {
  return 'identical(this, other) || ${_specificEqualityExpressionFor(type, body)}';
}

extension a on TypeParam {
  String toCode() => name.toCode() + ' extends ' + constraint.toCode();
}

extension aa on Iterable<TypeParam> {
  Iterable<TypeD> toInstantiation() => map((e) => TypeD(e.name));
  String toInstantiationCode() =>
      emptyOrSurrounded(toInstantiation().map((e) => e.toCode()));
  String toCode() => emptyOrSurrounded(map((e) => e.toCode()));
}

Iterable<String> assertStatementsFromTypeD(TypeD typeD, String name) =>
    typeD.assertions
        .map((e) => e.replaceAll('{}', name))
        .map((e) => 'assert($e)');

String argumentFromEntryToCode(MapEntry<Symbol, TypeD> e) =>
    '${e.value.toCode()} ${e.key.toCode()}';

String argumentFromEntryStringToCode(MapEntry<String, TypeD> e) =>
    '${e.value.toCode()} ${e.key}';

Iterable<String> namedArgumentsAndArgumentsToCode(
  Map<Symbol, TypeD> namedArguments,
  Map<Symbol, TypeD> requiredNamedArguments,
) =>
    namedArguments.entries.map(argumentFromEntryToCode).followedBy(
          requiredNamedArguments.entries
              .map(argumentFromEntryToCode)
              .map((e) => 'required $e'),
        );

String functionSignatureToCode(
  TypeD returnType,
  String name,
  Map<Symbol, TypeD> arguments, {
  Map<Symbol, TypeD> namedArguments = const {},
  Map<Symbol, TypeD> requiredNamedArguments = const {},
  Iterable<TypeParam> typeParams = const [],
}) =>
    '${returnType.toCode()} '
    '$name${typeParams.toCode()}'
    '('
    '${arguments.entries.map(argumentFromEntryToCode).join(', ')}'
    '${emptyOrSurrounded(namedArgumentsAndArgumentsToCode(namedArguments, requiredNamedArguments), '{,}')}'
    ')';

String functionTypeUnnamedToCode(
  TypeD returnType,
  Iterable<TypeD> arguments, [
  Iterable<TypeParam> typeParams = const [],
]) =>
    '${returnType.toCode()} '
    'Function'
    '${typeParams.toCode()}'
    '('
    '${arguments.map((e) => e.toCode()).join(', ')}'
    ')';

String functionTypeToCode(TypeD returnType, Map<Symbol, TypeD> arguments,
        [Iterable<TypeParam> typeParams = const []]) =>
    '${returnType.toCode()} '
    'Function'
    '${typeParams.toCode()}'
    '('
    '${arguments.entries.map(argumentFromEntryToCode).join(', ')}'
    ')';
bool _stringContains$(String s) => s.contains('\$');
String _escapedStringWith$(String s) => s.replaceAll(r'$', r'\$');
String _stringWithLeading$OrWrapped(String s) =>
    _stringContains$(s) ? '\${$s}' : '\$$s';
String _typeDToOutputString(TypeD type) =>
    '${_escapedStringWith$(type.name.toCode())}'
    '${emptyOrSurrounded(type.arguments.map((e) => _stringWithLeading$OrWrapped(e.toCode())))}';
String toStringFromTypeAndBody(String typePart, String bodyPart) =>
    '@override\n'
    'String toString() =>\n'
    ' "${bodyPart.isEmpty ? typePart : '$typePart $bodyPart'}"\n;';

String toStringFromTypeDAndBody(TypeD type, String bodyPart) =>
    toStringFromTypeAndBody(_typeDToOutputString(type), bodyPart);

String toStringBodyFrom(
  Iterable<Symbol> elements,
  String delimiterL,
  String delimiterR,
  bool outputOnEmpty,
) =>
    elements.isEmpty
        ? (outputOnEmpty ? '$delimiterL$delimiterR' : '')
        : '$delimiterL'
            '${elements.map((e) => _stringWithLeading$OrWrapped(e.toCode())).join(', ')}'
            '$delimiterR';

String copyWithSignatureToCode(TypeD classType, Map<Symbol, TypeD> elements) {
  final namedArgs =
      elements.map((key, value) => MapEntry(key, T(#Maybe, args: [value])));
  return '${classType.toCode()} '
      'copyWith'
      '('
      '${emptyOrSurrounded(namedArgs.entries.map(argumentFromEntryToCode).map((e) => "$e = const Maybe.none()"), '{,}')}'
      ')';
}

String copyWithToCode(TypeD classType, Map<Symbol, TypeD> elements,
        bool positionalArguments) =>
    """
  ${copyWithSignatureToCode(classType, elements)} => ${classType.toCode()}(
    ${elements.keys.map((e) => "${positionalArguments ? "${e.toCode()}: " : ""}${e.toCode()}.valueOr(this.${e.toCode()})").join(", ")}
  );
""";

String mixinToCode(List<TypeD> mixins) =>
    mixins.isEmpty ? '' : 'with ${mixins.map((e) => e.toCode()).join(',')}';

String toJsonSignatureToCode() =>
    functionSignatureToCode(TypeD.object, "toJson", {});

String toJsonObjectToCode(
  Map<Symbol, TypeD> body, [
  String? additionalExpression,
]) =>
    """
${toJsonSignatureToCode()} => {
  ${additionalExpression == null ? '' : '${additionalExpression},'}
  ${body.entries.map((e) => 'r"${e.key.toCode()}": ${_singleObjectToJsonToCode(e.key.toCode(), e.value)}').join(',')}
};
""";
String fromJsonObjectBody(
  String constructorName,
  Map<Symbol, TypeD> body,
  bool namedParameters,
) =>
    """
  $constructorName(
    ${body.entries.map((e) => '${namedParameters ? "${e.key.toCode()}: " : ""}${_singleObjectFromJsonToCode('(json as Map<String, Object?>)[r"${e.key.toCode()}"]', e.value)}').join(",")}
  )
""";

extension on Iterable<T> {
  Iterable<R> mapIndexed<R>(R Function(T, int) fn) sync* {
    var i = 0;
    for (final e in this) {
      yield fn(e, i);
      i++;
    }
  }
}

String _singleObjectFromJsonToCode(String expr, TypeD type) {
  if (type.fromJson == null) {
    return "$expr as ${type.toCode()}";
  }
  return "${type.fromJson}".replaceAll("{}", expr);
}

String _singleObjectToJsonToCode(String expr, TypeD type) {
  if (type.toJson == null) {
    return expr;
  }
  return "${type.toJson}".replaceAll("{}", expr);
}

String fromJsonListBody(
  Symbol className,
  List<TypeD> body,
) =>
    """
  ${className.toCode()}(
    ${body.mapIndexed((e, i) => _singleObjectFromJsonToCode("(json as List<Object?>)[$i]", e)).join(",")}
  )
""";

String toJsonListToCode(
  Map<Symbol, TypeD> body,
) =>
    """
${toJsonSignatureToCode()} => [
  ${body.entries.map((e) => _singleObjectToJsonToCode(e.key.toCode(), e.value)).join(',')}
];
""";
